from itertools import zip_longest
from copy import deepcopy


def pow_2(num: int) -> bool:
    """Checks if num is a power of two."""
    while num > 2:
        num /= 2
    if num == 2:
        return True
    return False


def non_1_or_0(num: int) -> bool:
    """Checks if there isn't a one or zero in num."""
    this = list(str(num))
    this = [int(i) for i in this]
    for i in this:
        if i > 1 or i < 0:
            return True
    return False


def dif_binary_types(x, y):
    """Checks if _binary types differ."""
    return x._binary is not y._binary


class BinaryError(Exception):
    """Raise if binary types differ, or if a non-one or non-zero is found."""
    pass


class BaseBits:
    """The base of both bits types."""
    def __init__(self, num, binary=False):
        assert num > -1, 'Binary object cannot be negative'
        self._num = num
        self._binary = bool(binary)
        if self._binary:
            if non_1_or_0(num):
                raise BinaryError('Non-one-or-zero found in Binary object')
            bits = list(reversed(list(str(self._num))))
            key = 0.5
            new_bits = {}
            for i in bits:
                key *= 2
                new_bits[int(key)] = int(i)
        else:
            self.encode()
            bits = list(reversed(list(str(self._num))))
            key = 0.5
            new_bits = {}
            for i in bits:
                key *= 2
                new_bits[int(key)] = int(i)
            self.decode()
        self._bits = [k for k, v in new_bits.items() if v == 1]

    def __repr__(self):
        return '%s(%s, binary=%s)' % (type(self).__name__,
                                      self._num,
                                      self._binary)

    def __str__(self):
        return str(self._num)

    def __format__(self, fmt_spec=''):
        if fmt_spec.endswith('c'):
            if not self._binary:
                self.encode()
                new_num = deepcopy(self._num)
                self.decode()
                fmt_spec = fmt_spec[:-1]
                return format(new_num, fmt_spec)
            fmt_spec = fmt_spec[:-1]
        elif fmt_spec.endswith('d'):
            if self._binary:
                self.decode()
                new_num = deepcopy(self._num)
                self.encode()
                fmt_spec = fmt_spec[:-1]
                return format(new_num, fmt_spec)
            fmt_spec = fmt_spec[:-1]
        elif fmt_spec.endswith('o'):
            if self._binary:
                self.decode()
                new_num = deepcopy(self._num)
                self.encode()
                fmt_spec = fmt_spec[:-1]
                return format(new_num, fmt_spec)
            else:
                self.encode()
                new_num = deepcopy(self._num)
                self.decode()
                fmt_spec = fmt_spec[:-1]
                return format(new_num, fmt_spec)
        return format(self._num, fmt_spec)

    def __len__(self):
        return len(self._bits)

    def __contains__(self, item):
        return item in self._bits

    def __getitem__(self, index):
        if isinstance(index, slice):
            return type(self)(sum(self._bits[index]))
        return self._bits[index]

    def __iter__(self):
        return (i for i in self._bits)

    def __add__(self, other):
        if isinstance(other, BaseBits):
            if dif_binary_types(self, other):
                msg = 'Binary types that differ cannot be added'
                raise BinaryError(msg)
            if not (self._binary and other._binary):
                return type(self)(self._num + other._num)
            self.decode()
            other.decode()
            this_num = deepcopy(self._num)
            that_num = deepcopy(other._num)
            self.encode()
            other.encode()
            result = type(self)(this_num + that_num)
            result.encode()
            return result
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, BaseBits):
            if dif_binary_types(self, other):
                msg = 'Binary types that differ cannot be subtracted'
                raise BinaryError(msg)
            if not (self._binary and other._binary):
                return type(self)(self._num - other._num)
            self.decode()
            other.decode()
            this_num = deepcopy(self._num)
            that_num = deepcopy(other._num)
            self.encode()
            other.encode()
            result = type(self)(this_num - that_num)
            result.encode()
            return result
        return NotImplemented

    def __mul__(self, scalar):
        if self._binary:
            self.decode()
            this_num = deepcopy(self._num)
            self.encode()
            result = type(self)(this_num * scalar)
            result.encode()
            return result
        return type(self)(self._num * scalar)

    def __rmul__(self, scalar):
        return self * scalar

    def __floordiv__(self, scalar):
        if self._binary:
            self.decode()
            this_num = deepcopy(self._num)
            self.encode()
            result = type(self)(this_num // scalar)
            result.encode()
            return result
        return type(self)(self._num // scalar)

    def __mod__(self, scalar):
        if self._binary:
            self.decode()
            this_num = deepcopy(self._num)
            self.encode()
            result = type(self)(this_num % scalar)
            result.encode()
            return result
        return type(self)(self._num % scalar)

    def __divmod__(self, other):
        return ((self // other)._num, (self % other)._num)

    def __gt__(self, other):
        if isinstance(other, BaseBits):
            if dif_binary_types(self, other):
                msg = 'Binary types that differ cannot be compared'
                raise BinaryError(msg)
            return bool(self._num > other._num)
        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, BaseBits):
            if dif_binary_types(self, other):
                msg = 'Binary types that differ cannot be compared'
                raise BinaryError(msg)
            return bool(self._num >= other._num)
        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, BaseBits):
            if dif_binary_types(self, other):
                msg = 'Binary types that differ cannot be compared'
                raise BinaryError(msg)
            return bool(self._num < other._num)
        return NotImplemented

    def __le__(self, other):
        if isinstance(other, BaseBits):
            if dif_binary_types(self, other):
                msg = 'Binary types that differ cannot be compared'
                raise BinaryError(msg)
            return bool(self._num <= other._num)
        return NotImplemented

    def __eq__(self, other):
        if isinstance(other, BaseBits):
            if dif_binary_types(self, other):
                msg = 'Binary types that differ cannot be compared'
                raise BinaryError(msg)
            return bool(self._num == other._num)
        return NotImplemented

    def encode(self):
        if not self._binary:
            self._binary = True
            self._num = int(format(self._num, 'b'))

    def decode(self):
        if self._num == 0 or self._num == 1:
            self._binary = False
        if self._binary:
            self._binary = False
            one = 1
            copy_self = list(reversed([int(num)
                                       for num in list(str(self._num))]))
            new_self = []
            for num in copy_self:
                if num == 1:
                    new_self.append(one)
                else:
                    new_self.append(0)
                one *= 2
            self._num = sum(new_self)


class FrozenBits(BaseBits):
    """Binary object that can encode and decode to and from base-two."""
    def __hash__(self):
        if self._binary:
            self.decode()
            return_val = deepcopy(self._num)
            self.encode()
            return hash(return_val)
        return hash(self._num)


class Bits(BaseBits):
    """Binary object that can encode and decode to and from base-two."""
    def __iadd__(self, other):
        self._bits = (self + other)._bits
        self._num = (self + other)._num
        return self

    def __isub__(self, other):
        self._bits = (self - other)._bits
        self._num = (self - other)._num
        return self

    def __imul__(self, scalar):
        self._bits = (self * scalar)._bits
        self._num = (self * scalar)._num
        return self

    def __ifloordiv__(self, scalar):
        self._bits = (self // scalar)._bits
        self._num = (self // scalar)._num
        return self

    def __imod__(self, scalar):
        self._bits = (self % scalar)._bits
        self._num = (self % scalar)._num
        return self
